/* Entry.java
 *
 * created: Sun Oct 11 1998
 *
 * This file is part of Artemis
 *
 * Copyright (C) 1998,1999  Genome Research Limited
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header: /nfs/disk222/yeastpub/Repository/powmap/diana/Entry.java,v 1.65 1999/06/21 15:54:23 kmr Exp $
 */

package diana;

import Diana;

import diana.sequence.*;

import uk.ac.sanger.pathogens.*;
import uk.ac.sanger.pathogens.embl.DocumentEntry;
import uk.ac.sanger.pathogens.embl.Range;
import uk.ac.sanger.pathogens.embl.Key;
import uk.ac.sanger.pathogens.embl.Location;
import uk.ac.sanger.pathogens.embl.LocationParseException;
import uk.ac.sanger.pathogens.embl.QualifierVector;
import uk.ac.sanger.pathogens.embl.InvalidKeyException;
import uk.ac.sanger.pathogens.embl.InvalidRelationException;

import java.util.NoSuchElementException;
import java.util.Vector;
import java.io.IOException;
import java.io.File;
import java.io.Reader;

/**
 *  This class is a wrapper for the embl.Entry class which contains the state
 *  and handles the events needed for editing one embl entry in Diana.  The
 *  state includes an EMBL.Entry object and a list of EntryChange listeners
 *  and FeatureChange listeners.  Other objects can register as listeners for
 *  changes to the entry.  For that reason, all changes to the embl.Entry
 *  objects should go through this class.  (see ChangeEvent for details of the
 *  possible events.)
 *
 *  @author Kim Rutherford
 *  @version $Id: Entry.java,v 1.65 1999/06/21 15:54:23 kmr Exp $
 **/

public class Entry implements FeatureChangeListener, Selectable {
  /**
   *  Create a new Entry object.
   *  @param bases The Bases object which contains the Strand objects that are
   *    used by the features of this Entry.
   *  @param embl_entry a reference to an embl.Entry object containing the
   *    underlying data for new object
   **/
  public Entry (final Bases bases,
                final uk.ac.sanger.pathogens.embl.Entry embl_entry) {
    this.embl_entry = embl_entry;
    this.bases = bases;

//     System.out.println ("Entry.Entry (): " + this + " " + entry_group);

    createDianaFeatures ();
  }

  /**
   *  Returns true if and only if this entry is read only.
   **/
  public boolean isReadOnly () {
    return getEMBLEntry ().isReadOnly ();
  }

  /**
   *  Save the changes to this Entry back to where the Entry came from.  If
   *  this object is read only a ReadOnlyException will be thrown.
   **/
  public void save ()
      throws IOException {
    Thread.yield ();
    getEMBLEntry ().save ();
  }

  /**
   *  Save the changes to this Entry back to where the Entry came from (like
   *  save ()) but without any diana extensions.  If this object is read only
   *  or doesn't support this a ReadOnlyException will be thrown.
   **/
  public void saveEMBLOnly ()
      throws IOException {
    if (getEMBLEntry () instanceof DocumentEntry) {
      final DocumentEntry document_entry = (DocumentEntry) getEMBLEntry ();

      document_entry.save (false);
    } else {
      throw new ReadOnlyException ("this save function cannot be " +
                                   "applied to this entry");
    }
  }

  /**
   *  Save the changes to this Entry to the file with the given name.  If this
   *  object is read only a ReadOnlyException will be thrown.  This method
   *  only works when the underlying embl.Entry object is a DocumentEntry.
   **/
  public void save (final String file_name) throws IOException {
    final FileDocument file_document =
      new FileDocument (new File (file_name));

    if (getEMBLEntry () instanceof DocumentEntry) {
      final DocumentEntry document_entry = (DocumentEntry) getEMBLEntry ();

      document_entry.save (file_document, true);
    } else {
      (new DocumentEntry (getEMBLEntry ())).save (file_document, true);
    }
  }

  /**
   *  Save the changes to this Entry to the file with the given name (like
   *  save (Document)) but without any diana extensions.  If this
   *  object is read only a ReadOnlyException will be thrown.  This method
   *  only works when the underlying embl.Entry object is a DocumentEntry.
   **/
  public void saveEMBLOnly (final String file_name) throws IOException {
    final FileDocument file_document =
      new FileDocument (new File (file_name));

    if (getEMBLEntry () instanceof DocumentEntry) {
      final DocumentEntry document_entry = (DocumentEntry) getEMBLEntry ();

      document_entry.save (file_document, false);
    } else {
      (new DocumentEntry (getEMBLEntry ())).save (file_document, false);
    }
  }

  /**
   *  Create a new Entry in the given EntryGroup.
   *  @param entry_group The new Entry will be added to this EntryGroup.
   *  @return A reference to the new Entry.
   **/
  public static Entry newEntry (final Bases bases) {
    final uk.ac.sanger.pathogens.embl.Entry entry = new DocumentEntry ();

    return new Entry (bases, entry);
  }

  /**
   *  Return the name of this Entry or null if it has no name.
   **/
  public String getName () {
    return getEMBLEntry ().getName ();
  }

  /**
   *  Set the name of this Entry - if possible (the return value will let the
   *  caller know).
   *  @return true if and only if the name was successfully set.  Not all
   *    Entry objects can change there name, so it is up to the calling
   *    function to check the return value.
   **/
  public boolean setName (final String name) {
    final String orig_entry_name = getName ();

    if (getEMBLEntry ().setName (name)) {
      if (orig_entry_name == null || ! name.equals (orig_entry_name)) {

        // now inform the listeners that the name has changed
        final EntryChangeEvent event =
          new EntryChangeEvent (name, this, EntryChangeEvent.NAME_CHANGED);

        fireAction (entry_listener_list, event);
      }
      return true;
    } else {
      return false;
    }
  }

  /**
   *  Return the text of the EMBL header of this Entry or null if there is no
   *  header.
   **/
  public String getHeaderText () {
    return getEMBLEntry ().getHeaderText ();
  }

  /**
   *  Set the header of this Entry to be the given text.
   *  @return true if and only if the header was successfully set.  Not all
   *    Entry objects can change their header, so it is up to the calling
   *    function to check the return value.
   *  @exception IOException thrown if there is a problem reading the header
   *    from the String - most likely ReadFormatException.
   **/
  public boolean setHeaderText (final String new_header)
      throws IOException {
    if (getEMBLEntry ().setHeaderText (new_header)) {
      final EntryChangeEvent event =
        new EntryChangeEvent (new_header,
                              this,
                              EntryChangeEvent.HEADER_CHANGED);

      fireAction (entry_listener_list, event);

      return true;
    } else {
      return false;
    }
  }

  /**
   *  Return the path of the directory that this entry is in.
   **/
  public Document getRootDocument () {
    if (((uk.ac.sanger.pathogens.embl.DocumentEntry) getEMBLEntry ()).getDocument () != null &&
        getEMBLEntry () instanceof uk.ac.sanger.pathogens.embl.DocumentEntry) { // XXX FIXME:
      return ((uk.ac.sanger.pathogens.embl.DocumentEntry) getEMBLEntry ()).getDocument ().getParent ();
    } else {
      return null;
    }
  }

  /**
   *  Return a vector containing the references of the Feature objects within
   *  the given range.
   *  @param range Return features that overlap this range - ie the start of
   *    the feature is less than or equal to the end of the range and the end
   *    of the feature is greater than or equal to the start of the range.
   *  @return The non-source key features of this feature table the are within
   *    the given range.  The returned object is a copy - changes will not
   *    effect the FeatureTable object itself.
   **/
  public FeatureVector getFeaturesInRange (Range range)
      throws OutOfRangeException {
    final FeatureVector return_features = new FeatureVector ();

    final uk.ac.sanger.pathogens.embl.FeatureVector embl_features =
      getEMBLEntry ().getFeaturesInRange (range);

    for (int i = 0 ; i < embl_features.size () ; ++i) {
      final Feature diana_feature = getFeatureOf (embl_features.elementAt (i));
      return_features.add (diana_feature);
    }

    return return_features;
  }

  /**
   *  Return a vector containing the references of the Feature objects in this
   *  Entry.
   *  @return The non-source key features of this Entry.  The
   *    returned object is a copy - changes will not effect the Entry
   *    object itself.
   **/
  public FeatureVector getAllFeatures () {
    final FeatureVector return_features = new FeatureVector ();

    final uk.ac.sanger.pathogens.embl.FeatureVector embl_features =
      getEMBLEntry ().getAllFeatures ();

    for (int i = 0 ; i < embl_features.size () ; ++i) {
      final Feature diana_feature = getFeatureOf (embl_features.elementAt (i));
      return_features.add (diana_feature);
    }

    return return_features;
  }

  /**
   *  Delete a Feature from this object and the underlying embl.Entry object.
   *  @param feature The Feature to delete.
   **/
  public void remove (final Feature feature)
      throws ReadOnlyException {
    synchronized (getBases ()) {
//         System.out.println ("removing: " + feature);
//         System.out.println ("removing: " + feature.getEmblFeature ());

      // get the embl.Feature object out of the diana.Feature object then
      // remove it
      if (!getEMBLEntry ().remove (feature.getEmblFeature ())) {
        throw new Error ("internal error - remove failed");
      }

//       System.out.println ("remove (): " + old_feature_index);

      // now inform the listeners that a deletion has occured
      final EntryChangeEvent event =
        new EntryChangeEvent (this, feature, EntryChangeEvent.FEATURE_DELETED);

      fireAction (entry_listener_list, event);

      feature.setEntry (null);
    }
  }

  /**
   *  Remove all the features in this Entry.
   **/
  public void removeAllFeatures ()
      throws ReadOnlyException {
    while (getFeatureCount () > 0) {
      // remove the first feature
      remove (getFeature (0));
    }
  }

  /**
   *  Add a Feature to the Entry.  The new feature will be inserted in order
   *  in the feature vector.  The features in the vector are ordered by the
   *  first base of each feature.
   *  @param new_feature The new feature to add.  It should not be a member
   *    of the Entry object.
   **/
  public void add (Feature new_feature)
      throws ReadOnlyException {
    synchronized (getBases ()) {
//      System.out.println ("add: " + new_feature);
      if (contains (new_feature)) {
        throw new Error ("internal error - tried to re-add a feature");
      }

      new_feature.setEntry (this);

      final uk.ac.sanger.pathogens.embl.Feature old_embl_reference =
        new_feature.getEmblFeature ();

      // this call will add the new_feature in the correct place in the
      // embl.Feature vector of the FeatureTable object
      final uk.ac.sanger.pathogens.embl.Feature new_embl_reference =
        getEMBLEntry ().add (new_feature.getEmblFeature ());

      if (new_embl_reference != old_embl_reference) {
        new_feature.getEmblFeature ().setUserData (null);
        new_feature.setEmblFeature (new_embl_reference);
        new_embl_reference.setUserData (new_feature);
      }

      // now inform the listeners that a addition has occured
      final EntryChangeEvent event =
        new EntryChangeEvent (this, new_feature,
                              EntryChangeEvent.FEATURE_ADDED);

      fireAction (entry_listener_list, event);
    }
  }

  /**
   *  Forget all the features in this entry and inform all the EntryChange
   *  listeners that all the features have gone away.  The garbage collector
   *  will really dispose of the features later.
   **/
  public void dispose () {
    final FeatureEnumeration feature_enum = features ();

    while (feature_enum.hasMoreFeatures ()) {
      final Feature current_feature = feature_enum.nextFeature ();

      // now inform the listeners that a deletion has occured
      final EntryChangeEvent event =
        new EntryChangeEvent (this, current_feature,
                              EntryChangeEvent.FEATURE_DELETED);

      fireAction (entry_listener_list, event);

      current_feature.setEntry (null);
    }
  }

  /**
   *  Returns the Feature at the given index in the vector of Features.
   **/
  public Feature getFeature (int i) {
//    System.out.println (getEmblFeatures ().elementAt (i).getUserData ());

    return getFeatureOf (getEMBLEntry ().getFeatureAtIndex (i));
  }

  /**
   *  Create a new Feature in this Entry with no qualifiers, a key of
   *  "misc_feature" and a location of "1"
   **/
  public Feature createFeature ()
      throws ReadOnlyException {
    final Key new_key;
    final Location new_location;
    final QualifierVector new_qualifiers = new QualifierVector ();

    try {
      new_key = new Key ("misc_feature");
      new_location = new Location ("1");

      return createFeature (new_key, new_location, new_qualifiers);
    } catch (LocationParseException e) {
      throw new Error ("internal error - unexpected exception: " + e);
    } catch (InvalidKeyException e) {
      throw new Error ("internal error - unexpected exception: " + e);
    } catch (InvalidRelationException e) {
      throw new Error ("internal error - unexpected exception: " + e);
    } catch (OutOfRangeException e) {
      throw new Error ("internal error - unexpected exception: " + e);
    }
  }

  /**
   *  Create and return a new Feature in this Entry with the given key,
   *  location and qualifiers.  This method creates a new embl.Feature object
   *  and then wraps it in a diana.Feature object and then inserts it in this
   *  Entry object.
   *  @param new_key The Key to use for the new Feature.
   *  @param new_location The Location to use for the new Feature.
   *  @param new_qualifiers The a vector containing the Qualifier objects to
   *    use for the new Feature.
   *  @exception InvalidRelationException Thrown if this Feature cannot
   *    contain the given Qualifier.
   **/
  public Feature createFeature (Key new_key,
                                Location new_location,
                                QualifierVector new_qualifiers)
      throws InvalidRelationException, ReadOnlyException, OutOfRangeException {
    final uk.ac.sanger.pathogens.embl.Feature new_embl_feature =
      getEMBLEntry ().createFeature (new_key, new_location, new_qualifiers);

    final Feature new_feature = new Feature (this, new_embl_feature);

    // now inform the listeners that a addition has occured
    final EntryChangeEvent event =
      new EntryChangeEvent (this, new_feature, EntryChangeEvent.FEATURE_ADDED);

    fireAction (entry_listener_list, event);

    return new_feature;
  }

  /**
   *  Create and return a new Feature in this Entry with the given key,
   *  location and no qualifiers.  This method creates a new embl.Feature
   *  object and then wraps it in a diana.Feature object and then inserts it
   *  in this Entry object.
   *  @param new_key The Key to use for the new Feature.
   *  @param new_location The Location to use for the new Feature.
   **/
  public Feature createFeature (Key new_key,
                                Location new_location)
      throws ReadOnlyException, OutOfRangeException {
    try {
      final QualifierVector new_qualifiers = new QualifierVector ();

      return createFeature (new_key, new_location, new_qualifiers);
    } catch (InvalidRelationException e) {
      throw new Error ("internal error - unexpected exception: " + e);
    }
  }

  /**
   *  Return true if this Entry contains the given Feature.
   **/
  public boolean contains (Feature feature) {
    if (indexOf (feature) == -1) {
      return false;
    } else {
      return true;
    }
  }

  /**
   *  Return the index of the given Feature in the feature table of this Entry
   *  or -1 if the feature isn't in this Entry.
   **/
  public int indexOf (Feature feature) {
    return getEMBLEntry ().indexOf (feature.getEmblFeature ());
  }

  /**
   *  Return the index of a feature in the embl feature table of this entry.
   *  The indices start at zero.
   *  @param feature The Feature to lookup.
   *  @return The index of the feature or -1 if the feature isn't in this
   *    Entry.
   **/
  public int getIndexOfFeature (Feature feature) {
    return getEMBLEntry ().indexOf (feature.getEmblFeature ());
  }

  /**
   *  Returns an enumeration of the Feature objects in this EntryGroup. The
   *  returned FeatureEnumeration object will generate all features in this
   *  object in turn. The first item generated is the item at index 0, then
   *  the item at index 1, and so on.
   **/
  public FeatureEnumeration features () {
//    System.out.println ("features (): " + this);
    return new FeatureEnumerator ();
  }

  /**
   *  An Enumeration of Feature objects.
   **/
  public class FeatureEnumerator implements FeatureEnumeration {
    /**
     *  Create a new FeatureEnumeration that will enumerate the enclosing
     *  Entry object.  The Entry object must not be changed while the
     *  enumeration is active.
     **/
    public FeatureEnumerator () {
      feature_enumerator = getEMBLEntry ().features ();
    }

    /**
     *  See the FeatureEnumeration interface for details.
     **/
    public boolean hasMoreFeatures () {
      return feature_enumerator.hasMoreFeatures ();
    }

    /**
     *  See the FeatureEnumeration interface for details.
     **/
    public Feature nextFeature ()
        throws NoSuchElementException {

      final uk.ac.sanger.pathogens.embl.Feature this_feature =
        feature_enumerator.nextFeature ();

      return getFeatureOf (this_feature);
    }

    /**
     *  The Enumeration for the current entry
     **/
    private uk.ac.sanger.pathogens.embl.FeatureEnumeration feature_enumerator;
  }

  /**
   *  Returns the number of features in this entry.
   **/
  public int getFeatureCount () {
    return getEMBLEntry ().getFeatureCount ();
  }

  /**
   *  Return the base length of the sequence of this entry.
   **/
  private int getSequenceLength () {
    return getBases ().getLength ();
  }

  /**
   *  Return the object representing the forward sequence of bases for this
   *  Entry.
   **/
  private Strand getForwardStrand () {
    return getBases ().getForwardStrand ();
  }

  /**
   *  Return the object representing the backward (reverse complemented)
   *  sequence of bases for this Entry.
   **/
  private Strand getBackwardStrand () {
    return getBases ().getBackwardStrand ();
  }

  /**
   *  Return the Bases object that was passed to the constructor.
   **/
  public Bases getBases () {
    return bases;
  }

  /**
   *  Returns a vector containing those CDS features that have no /pseudo
   *  qualifier and do not have valid start and stop codons.  The returned
   *  features may be illegal in EMBL submissions.
   **/
  public FeatureVector checkForStopsAndStarts () {
    // get all the CDS features that do not have a /pseudo qualifier
    final FeatureKeyQualifierPredicate predicate =
      new FeatureKeyQualifierPredicate (Key.CDS, "pseudo", false);

    final String start_codons = Diana.getOptions ().getStartCodonSymbols ();

    final FeatureVector non_embl_features = new FeatureVector ();

    final FeatureEnumeration feature_enum = features ();

    while (feature_enum.hasMoreFeatures ()) {
      final Feature current_feature = feature_enum.nextFeature ();

      if (predicate.testPredicate (current_feature)) {
        final AminoAcidSequence translation =
          current_feature.getTranslation ();

        if (translation.length () < 1) {
          non_embl_features.add (current_feature);
          continue;
        }

        final char first_codon = translation.elementAt (0);

        final int codon_start_index = current_feature.getCodonStart () - 1;

        final String feature_bases =
          current_feature.getBases ().substring (codon_start_index);

        if (feature_bases.length () % 3 != 0) {
          non_embl_features.add (current_feature);
          continue;
        }

        final String codon_string =
          feature_bases.substring (feature_bases.length () - 3);

        final char last_codon =
          AminoAcidSequence.getCodonTranslation (codon_string);

        if (start_codons.indexOf (first_codon) == -1) {
          non_embl_features.add (current_feature);
          continue;
        }

        if (!AminoAcidSequence.isStopCodon (last_codon)) {
          non_embl_features.add (current_feature);
          continue;
        }
      }
    }

    return non_embl_features;
  }

  /**
   *  Returns a vector containing those features that have the same key and
   *  location as one or more other features.  These features are not allowed
   *  in EMBL submissions.
   **/
  public FeatureVector checkForEMBLDuplicates () {
    final FeatureVector non_embl_features = new FeatureVector ();

    final FeatureEnumeration feature_enum = features ();

    Feature last_feature = null;

    while (feature_enum.hasMoreFeatures ()) {

      final Feature current_feature = feature_enum.nextFeature ();

      final Key current_key = current_feature.getKey ();

      final Location current_location = current_feature.getLocation ();

      if (last_feature != null &&
          last_feature.getKey ().equals (current_key) &&
          last_feature.getLocation ().equals (current_location)) {
        non_embl_features.add (last_feature);
        non_embl_features.add (current_feature);
        continue;
      }

      last_feature = current_feature;
    }

    return non_embl_features;
  }

  /**
   *  Reverse and complement the sequence and all features in this Entry.
   **/
  public void reverseComplement ()
      throws ReadOnlyException {
    // save the references of all the features
    final FeatureVector all_features = getAllFeatures ();

    // remove all the Feature objects, change their location and then add them
    // back.  we do this because the embl.FeatureTable depends on having a
    // consistent ordering of embl.Feature objects, which isn't the case
    // during this procedure.
    for (int i = 0 ; i < all_features.size () ; ++i) {
      remove (all_features.elementAt (i));
    }

    for (int i = 0 ; i < all_features.size () ; ++i) {
      final Feature this_feature = all_features.elementAt (i);

      this_feature.updateEMBLFeatureLocation ();
      add (this_feature);

      // now inform the listeners that the location of the feature has changed
      final FeatureChangeEvent event =
        new FeatureChangeEvent (this,
                                this_feature,
                                FeatureChangeEvent.LOCATION_CHANGED);

//      featureChanged (event);
    }
  }

  /**
   *  Adds the specified event listener to receive entry change events from
   *  this object.
   *  @param l the event change listener.
   **/
  public void addEntryChangeListener (EntryChangeListener l) {
    entry_listener_list.addElement (l);
  }

  /**
   *  Removes the specified event listener so that it no longer receives
   *  entry change events from this object.
   *  @param l the event change listener.
   **/
  public void removeEntryChangeListener (EntryChangeListener l) {
    entry_listener_list.removeElement (l);
  }

  /**
   *  Adds the specified event listener to receive feature change events from
   *  this object.
   *  @param l the event change listener.
   **/
  public void addFeatureChangeListener (FeatureChangeListener l) {
    feature_listener_list.addElement (l);
  }

  /**
   *  Removes the specified event listener so that it no longer receives
   *  feature change events from this object.
   *  @param l the event change listener.
   **/
  public void removeFeatureChangeListener (FeatureChangeListener l) {
    feature_listener_list.removeElement (l);
  }

  /**
   *  Implementation of the FeatureChangeListener interface.  We need to
   *  listen to feature change events from the Features in this object.
   *  @param event The change event.
   **/
  public void featureChanged (FeatureChangeEvent event) {
//     System.out.println ("Entry.featureChanged () - feature change event: " +
//                         event.getType () + " for feature " +
//                         event.getSource ());

    // pass the action straight through
    fireAction (feature_listener_list, event);
  }

  /**
   *  Send an event to those object listening for it.
   *  @param listeners A Vector of the objects that the event should be sent
   *    to.
   *  @param event The event to send
   **/
  private void fireAction (Vector listeners, ChangeEvent event) {
    final Vector targets;
    // copied from a book - synchronising the whole method might cause a
    // deadlock
    synchronized (this) {
      targets = (Vector) listeners.clone ();
    }

    for (int i = 0 ; i < targets.size () ; ++i) {
      ChangeListener target = (ChangeListener) targets.elementAt (i);

      if (event instanceof EntryChangeEvent) {
        final EntryChangeListener entry_change_listener =
          (EntryChangeListener) target;
        entry_change_listener.entryChanged ((EntryChangeEvent) event);
      } else {
        final FeatureChangeListener feature_change_listener =
          (FeatureChangeListener) target;
        feature_change_listener.featureChanged ((FeatureChangeEvent) event);
      }
    }
  }

  /**
   *  Return the diana.Feature object of the given embl.Feature object.  This
   *  method will create an appropriate diana.Feature if none exists.
   **/
  private Feature getFeatureOf (uk.ac.sanger.pathogens.embl.Feature embl_feature) {
    final Feature test_feature = (Feature) embl_feature.getUserData ();

    if (test_feature == null) {
      final Feature new_feature = new Feature (this, embl_feature);
      return new_feature;
    } else {
      return test_feature;
    }
  }

  /**
   *  Create a diana.Feature object for each embl.Feature in embl_entry.
   **/
  private void createDianaFeatures () {
    // enumerating the features will call getFeatureOf () for each
    // embl.Feature, which will in turn create a diana.Feature

    final FeatureEnumeration feature_enumerator = features ();

    while (feature_enumerator.hasMoreFeatures ()) {
      final Feature this_feature = feature_enumerator.nextFeature ();
    }
  }

  /**
   *  Return the embl.Entry object that was passed to the constructor.
   **/
  uk.ac.sanger.pathogens.embl.Entry getEMBLEntry () {
    return embl_entry;
  }

  /**
   *  The embl.Entry object that was passed to the constructor
   **/
  private uk.ac.sanger.pathogens.embl.Entry embl_entry;

  /**
   *  A vector of those objects listening for entry change events.
   **/
  final private Vector entry_listener_list = new Vector ();

  /**
   *  A vector of those objects listening for feature change events.
   **/
  final private Vector feature_listener_list = new Vector ();

  /**
   *  This is the Bases reference that was passed to the constructor.
   **/
  /*final*/ private Bases bases;
}
