/* FeatureSegment.java
 *
 * created: Sun Oct 11 1998
 *
 * This file is part of Artemis
 * 
 * Copyright (C) 1998,1999  Genome Research Limited
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header: /nfs/disk222/yeastpub/Repository/powmap/diana/FeatureSegment.java,v 1.33 1999/06/29 14:05:24 kmr Exp $
 */

package diana;

import uk.ac.sanger.pathogens.*;
import uk.ac.sanger.pathogens.embl.Range;
import uk.ac.sanger.pathogens.embl.FuzzyRange;
import uk.ac.sanger.pathogens.embl.Key;
import uk.ac.sanger.pathogens.embl.Location;
import uk.ac.sanger.pathogens.embl.LocationParseNode;
import uk.ac.sanger.pathogens.embl.LocationParseNodeVector;
import diana.sequence.*;
import diana.components.FeatureDisplay;

import java.util.Vector;

/**
 *  Objects of this class will generally represent one exon.  We use the name
 *  "segment" because EMBL does not specify that a feature with a join or order
 *  must be a coding sequence.
 *
 *  @author Kim Rutherford
 *  @version $Id: FeatureSegment.java,v 1.33 1999/06/29 14:05:24 kmr Exp $
 *
 **/

public class FeatureSegment implements Selectable, MarkerChangeListener {
  /**
   *  Create a new FeatureSegment object.
   *  @param feature The reference of the Feature that contains this segment.
   *  @param range The raw embl.Range of this segment.
   **/
  public FeatureSegment (Feature feature, Range range) {
    this.feature      = feature;
    this.range        = range;
    
    final int start_position;

    final int end_position;

    if (getFeature ().isForwardFeature ()) {
      start_position = range.getStart ();
      end_position = range.getEnd ();
    } else {
      // markers are reversed on the reverse strand
      start_position = range.getEnd ();
      end_position = range.getStart ();
    }
    
    final Strand strand = getFeature ().getStrand ();

    try {
      start = strand.makeMarkerFromRawPosition (start_position);
      end   = strand.makeMarkerFromRawPosition (end_position);

    } catch (OutOfRangeException e) {
      throw new Error ("internal error - " +
                       "unexpected OutOfRangeException for position: " +
                       start_position);
    }
  }

  /**
   *  Return the owning Feature of this segment (as passed to the constructor).
   **/
  public Feature getFeature () {
    return feature;
  }

  /**
   *  Return the start position Marker of this object.
   **/
  public Marker getStart () {
    return start;
  }

  /**
   *  Return the end position Marker of this object.
   **/
  public Marker getEnd () {
    return end;
  }

  /**
   *  Set the end Marker of this segment.
   **/
  void setEnd (final Marker new_end) {
    end = new_end;
  }

  /**
   *  Return the range (in bases of this segment).
   **/
  public Range getRange () {
    return range;
  }

  /**
   *  Remove this FeatureSegment from the Feature that contains it.
   **/
  public void removeFromFeature () {
    System.out.println ("FeatureSegment.removeFromFeature (): not implemented");
//      getFeature ().removeSegment (this);
//      feature = null;
  }
  
  /**
   *  Return the number of bases in this feature (total of all segments).
   **/
  public int getBaseCount () {
    return getEnd ().getPosition () - getStart ().getPosition () + 1;
  }

  /**
   *  Return the bases in this FeatureSegment.
   **/
  public String getBases () {
    final Strand strand = getFeature ().getStrand ();
    try {
      return strand.getSubSequence (new Range (getStart ().getPosition (),
                                               getEnd ().getPosition ()));
    } catch (OutOfRangeException e) {
      throw new Error ("internal error - unexpected exception: " + e);
    }
  }

  /**
   *  Return true if and only if the Marker objects at the ends of this
   *  segment can be changed.
   **/
  public boolean canDirectEdit () {
    return ! (range instanceof FuzzyRange);
  }
  
  /**
   *  Return the LocationParseNode of this feature's Location that contains
   *  the Range of this segment.
   **/
  private LocationParseNode findRangeNode () {
    final Location location = getFeature ().getEmblFeature ().getLocation ();

    final LocationParseNode parsed_location = location.getParsedLocation ();

    final LocationParseNode return_node;

    final LocationParseNodeVector search_nodes;

    if (parsed_location.getType () == LocationParseNode.JOIN) {
      search_nodes = parsed_location.getJoinChildren ();
    } else {
      if (parsed_location.getType () == LocationParseNode.ORDER) {
        search_nodes = parsed_location.getOrderChildren ();
      } else {
        search_nodes = new LocationParseNodeVector ();
        search_nodes.addElement (parsed_location);
      }
    }

    for (int i = 0 ; i < search_nodes.size () ; ++i) {
      final LocationParseNode current_node = search_nodes.elementAt (i);

      if (current_node.getType () == LocationParseNode.COMPLEMENT) {
        if (current_node.getComplementChild ().getRange ().equals (range)) {
          return current_node.getComplementChild ();
        }
      } else {
        if (current_node.getRange ().equals (range)) {
          return current_node;
        }
      }
    }

    throw new Error ("internal error - Range not found");
  }

  /**
   *  Implementation of the MarkerChangeListener interface.  We get the
   *  embl.Location object from the feature and then "fix" it and put it back.
   **/
  public void markerChanged (final MarkerChangeEvent event) {
    // find the node that refers to this segment's range
    final LocationParseNode range_node = findRangeNode ();

    final Range new_range;

    if (getFeature ().isForwardFeature ()) {

      // this is a big hack - if some bases are added or deleted from the
      // sequence both the start and end Markers will get events separately,
      // so the start Marker may be temporarily be after the end Marker - in
      // that case we ignore the event.
      if (getStart ().getRawPosition () > getEnd ().getRawPosition ()) {
        return;
      }

      new_range = range.change (getStart ().getRawPosition (),
                                getEnd ().getRawPosition ());
    } else {
      // see the comment above
      if (getEnd ().getRawPosition () > getStart ().getRawPosition ()) {
        return;
      }

      new_range = range.change (getEnd ().getRawPosition (),
                                 getStart ().getRawPosition ());
    }

    // Range.change () will return null if the Marker can't be changed just by
    // setting the start and end positions (eg. FuzzyRange)

    if (new_range != null && !new_range.equals (range)) {
      try {
        final uk.ac.sanger.pathogens.embl.Entry entry =
          getFeature ().getEntry ().getEMBLEntry ();

        // we need to remove the feature while changing it's location
        entry.remove (getFeature ().getEmblFeature ());

        range_node.setRange (new_range);

        range = new_range;

        // re-insert this feature into the entry so that it has the correct
        // position in the entry
        entry.add (getFeature ().getEmblFeature ());

        getFeature ().locationChanged ();
      } catch (ReadOnlyException e) {
        // just ignore the change
      }
    }
  }

  /**
   *  Adds the specified event listener to receive marker change events from
   *  this object.  An event will be sent when either the start or end Marker
   *  of this segment changes.
   *  @param l the change event listener.
   **/
  public void addMarkerChangeListener (MarkerChangeListener l) {
//    System.out.println ("addMarkerChangeListener () 1");
    marker_listener_list.addElement (l);
    if (marker_listener_list.size () == 1) {
      // we need to start listening
      start.addMarkerChangeListener (this);
      end.addMarkerChangeListener (this);
    }
  }

  /**
   *  Removes the specified event listener so that it no longer receives
   *  marker change events from this object. 
   *  @param l the change event listener.
   **/
  public void removeMarkerChangeListener (MarkerChangeListener l) {
//    System.out.println ("removeMarkerChangeListener () 1");
    marker_listener_list.removeElement (l);
    if (marker_listener_list.size () == 0) {
      // stop listening
      start.removeMarkerChangeListener (this);
      end.removeMarkerChangeListener (this);
    }
  }

  /**
   *  Send a MarkerChangeEvent to those object listening for it.
   *  @param event The event to send
   **/
  private void fireEvent (MarkerChangeEvent event) {
    final Vector targets;
    // copied from a book - synchronising the whole method might cause a
    // deadlock
    synchronized (this) {
      targets = (Vector) marker_listener_list.clone ();
    }

    for (int i = 0 ; i < targets.size () ; ++i) {
      final MarkerChangeListener target =
        (MarkerChangeListener) targets.elementAt (i);

      target.markerChanged (event);
    }
  }

  /**
   *  Duplicate the segment and insert the new segment into the feature.
   **/
  public void duplicate () {
    // XXX
  }

  /**
   *  Returns 0, 1 or 2 depending on which translation frame this segment is
   *  in.  A frame shift of zero means that the bases should be translated
   *  starting at the start position of this segment, 1 means start
   *  translating one base ahead of the start position and 2 means start
   *  translating two bases ahead of the start position.
   **/
  private int getFrameShift () {
//    System.out.print ("getFrameShift (): " + frame_shift);
    
    // find the number of bases in the segments before this one
    int base_count = 0;

    final FeatureSegmentVector segments =
      getFeature ().getSegments ();
    
    for (int i = 0 ; i < segments.size () ; ++i) {
      if (segments.elementAt (i) == this) {
        break;
      } else {
        base_count += segments.elementAt (i).getBaseCount ();
      }
    }

    int foo = (base_count - (getFeature ().getCodonStart () - 1)) % 3;
      
//    System.out.println ("   " + foo);
    
    return foo;
  }

  /**
   *  Return the frame ID of the frame that this feature.  There are three
   *  forward and three reverse frame lines as well a forward strand and a
   *  backward strand line.  This method returns an ID of the line to place
   *  this segment on.  Normally only CDS and other protein features are
   *  placed on the frame lines (the rest go on the forward or backward strand
   *  line).
   *  @param frame_features_flag All features (not just CDS features) will be
   *    put on the frame lines if and only if this argument is true.
   **/
  public int getFrameID (boolean frame_features_flag) {
    boolean use_frame = frame_features_flag;

    if (getFeature ().isProteinFeature ()) {
      use_frame = true;
    }

    // this will be 0, 1 or 2 depending on which frame the segment is in
    final int start_base_modulo =
      (getStart ().getPosition () - 1 + 3 - getFrameShift ()) % 3;

    if (getFeature ().getStrand ().isForwardStrand ()) {
      if (use_frame) {
        switch (start_base_modulo) {
        case 0:
          return FeatureDisplay.FORWARD_FRAME_1;
        case 1:
          return FeatureDisplay.FORWARD_FRAME_2;
        case 2:
          return FeatureDisplay.FORWARD_FRAME_3;
        }
      } else {
        return FeatureDisplay.FORWARD_STRAND;
      }
    } else {
      if (use_frame) {
        switch (start_base_modulo) {
        case 0:
          return FeatureDisplay.BACKWARD_FRAME_1;
        case 1:
          return FeatureDisplay.BACKWARD_FRAME_2;
        case 2:
          return FeatureDisplay.BACKWARD_FRAME_3;
        }
      } else {
        return FeatureDisplay.BACKWARD_STRAND;
      }
    }

    return FeatureDisplay.NO_FRAME;
  }

  /**
   *  The reference of the Feature that contains this segment, as passed to
   *  the constructor.
   **/
  private Feature feature;

  /**
   *  The start position of the segment, set in getStart ().
   **/
  private Marker start = null;

  /**
   *  The end position of the segment, set in getEnd ().
   **/
  private Marker end = null;

  /**
   *  This is the Range reference that was passed to the constructor.
   **/
  private Range range;

  /**
   *  A vector containing the references of those objects listening for
   *  marker change events.
   **/
  private final Vector marker_listener_list = new Vector ();
}
